@@ -145,9 +145,15 @@ class Test_UrlIterator(unittest.TestCase):
         )
 
 
+class HTMLParserException(Exception):
+    pass
+
+
 class HTMLParser(object):
 
     def __init__(self, body_unicode):
+        if body_unicode.strip() == '':
+            raise HTMLParserException('Body empty.')
         htmlparser = lxml.etree.HTMLParser()
         self._body_unicode = self._prepare_body(body_unicode)
         #
@@ -156,15 +162,19 @@ class HTMLParser(object):
         try:
             self._tree = lxml.etree.fromstring(self._body_unicode, htmlparser)
         except:
-            print body_unicode
+            print '>>>>>>>>>>>>>>>> lxml.etree.fromstring failed', len(body_unicode)
+            print body_unicode.encode('utf-8')
             raise
+        if self._tree is None:
+            errmsg = 'HTML parsing failed'
+            raise HTMLParserException(errmsg)
 
     def _prepare_body(self, body_unicode):
         #
         # Hairy fix to remove encoding='..' from xml headers, as
-        # lxml doesn't accept these.
+        # lxml doesn't accept those.
         #
-        m = re.match(r'''(?i)<\?xml[\s\w'=.]+(encoding=['"][^'"]*['"])''', body_unicode)
+        m = re.match(r'''(?i)\s*<\?xml[\s\w'"=.]+(encoding=['"][^'"]*['"])''', body_unicode)
         if m:
             body_unicode = ''.join(body_unicode.split(m.group(1)))
         return body_unicode
