@@ -415,10 +415,9 @@ class Tool(object):
     def error(self):
         return self._error
 
-    def set_error(self, error, oneword=False, verbosity_limit=0):
-        if oneword:
-            error = str(error).lower()
-            error = error.replace(' ', '-')
+    def set_error(self, error, verbosity_limit=0):
+        error = str(error).lower()
+        error = error.replace(' ', '-')
         self._error = error
         self._error_verbosity_limit = verbosity_limit
 
@@ -583,6 +582,7 @@ class ToolMixin_AddrIterator(object):
         result = {'addr': addr}
         return result
 
+
 class VerbosityPrinter:
     def __init__(self, options, outfile=sys.stdout):
         assert hasattr(options, 'verbose')
@@ -598,7 +598,6 @@ class VerbosityPrinter:
             if len(argstr):
                 argstr = ' ' + argstr
             msg += argstr
-        msg += '\n'
         self._outfile.write(msg)
 
 
