@@ -120,7 +120,7 @@ class NetcatsTool(Tool):
             errmsg = e.strerror.lower()
             self.set_error(errmsg)
             return
-        self.vprint_stderr(1, '[connected to {0}:{1}]'.format(*addr))
+        self.vprint_stderr(1, '[connected to {0}:{1}]\n'.format(*addr))
 
         outfile = sys.stdout
         if self.options.split_output:
@@ -192,7 +192,7 @@ class NetcatsTool(Tool):
                     action_result['received'],
                 )
         host = action_result['addr']
-        self.vprint_stderr(1, '[connection closed {0}:{1}]'.format(*host))
+        self.vprint_stderr(1, '[connection closed {0}:{1}]\n'.format(*host))
         return output
 
     def __timeout__(self):
