@@ -13,6 +13,7 @@ import itertools
 import time
 import fcntl
 import subprocess
+import signal
 
 import gevent
 import gevent.pool
@@ -20,6 +21,7 @@ import gevent.monkey
 from gevent.socket import wait_read
 
 import xnet.tools
+from xnet.tools import VerbosityPrinter
 from xnet.net.ssh import SSHPopen
 
 
@@ -38,6 +40,7 @@ if DEVELOPMENT:
 
 DEFAULT_NR_MICROTHREADS='128'
 
+
 class XNetOptionParser(optparse.OptionParser):
     '''
         Provides some common options and an interface to add nicely formatted
@@ -102,6 +105,43 @@ class XNetOptionParser(optparse.OptionParser):
             self.add_option(short_, long_, help=help_, **kw)
 
 
+class SignalHandler(object):
+    __shared_state = {}
+
+    @classmethod
+    def setup(cls, options):
+        cls._options = options
+        cls.__shared_state.setdefault(
+            'vprint_stderr', VerbosityPrinter(options, sys.stderr)
+        )
+
+    def __init__(self):
+        self.__dict__ = self.__shared_state
+        self.__dict__.setdefault('_cleanups', [])
+        self.__dict__.setdefault('_cleanups', [])
+
+    @classmethod
+    def _signal_handler(cls):
+        cls.__class__().do_cleanup()
+
+    def do_cleanup(self):
+        self.vprint_stderr(0, '[*] ii do_cleanup()\n')
+        for (priority, fn, args, kw, _msg) in sorted(self._cleanups):
+            if not _msg is None:
+                self.vprint_stderr(0, '[*] {0}\n'.format(_msg))
+            fn(*args, **kw)
+
+    def register_cleanup(self, priority, fn, _msg=None, *args, **kw):
+        self._cleanups.append((priority, fn, args, kw, _msg))
+
+    def activate(self):
+        signal.signal(signal.SIGINT, SignalHandler._signal_handler)
+
+
+
+
+
+
 def get_action_usage():
     _action_usage = ''
     for tup in _available_actions:
@@ -197,18 +237,28 @@ def repeaterator(it, count):
             count -= 1
 
 
-def do_ssh(opt, ToolImplementation, args, inputchain):
+def do_ssh(options, ToolImplementation, args, inputchain):
     argv = sys.argv[:]
     subprocesses = []
-    for ssh_node in SSHPopen.parse_ssh_nodes_file(opt.ssh_nodes_file):
+    sighandler = SignalHandler()
+    vprint_stderr = VerbosityPrinter(options, sys.stderr)
+    #
+    for ssh_node in SSHPopen.parse_ssh_nodes_file(options.ssh_nodes_file):
         user, host = ssh_node.split('@', 1)
         env = {}
-        if opt.ssh_display:
-            env['DISPLAY'] = opt.ssh_display
-        ssh_args = opt.ssh_args or []
+        if options.ssh_display:
+            env['DISPLAY'] = options.ssh_display
+        ssh_args = options.ssh_args or []
         proc = SSHPopen(ssh_args, argv, args, user=user, host=host,
                         stdin=subprocess.PIPE, env=env)
         subprocesses.append(proc)
+        #
+        # Register cleanup handlers in case of ^C etc.
+        #
+        cleanup_msg = 'SSHPopen.pid {0}, stdin.close()'.format(proc.pid)
+        sighandler.register_cleanup(0, proc.stdin.close, _msg=cleanup_msg)
+        cleanup_msg = 'SSHPopen.pid {0}, wait()'.format(proc.pid)
+        sighandler.register_cleanup(100, proc.wait, _msg=cleanup_msg)
 
     for (i, item) in enumerate(inputchain):
         #
@@ -223,16 +273,16 @@ def do_ssh(opt, ToolImplementation, args, inputchain):
         proc.stdin.write(item)
 
     for proc in subprocesses:
-        #print os.getpid(), 'SSH-CLOSE-STDIN', proc
+        vprint_stderr('[*] ii', 'SSH-CLOSE-STDIN', os.getpid(), proc)
         proc.stdin.close()
 
     for proc in subprocesses:
-        #print os.getpid(), 'SSH-WAIT', proc
+        vprint_stderr('[*] ii', 'SSH-WAIT', os.getpid(), proc)
         proc.wait()
     return 0
 
 
-def do_fork(opt, ToolImplementation, tool_reads_stdin, args, nr_processes, nr_microthreads, inputchain):
+def do_fork(options, ToolImplementation, tool_reads_stdin, args, nr_processes, nr_microthreads, inputchain):
     #
     # Remove --nr-processes from sys.argv and spawn children.
     #
@@ -248,6 +298,7 @@ def do_fork(opt, ToolImplementation, tool_reads_stdin, args, nr_processes, nr_mi
     if '--nr-processes={0}'.format(nr_processes) in argv:
         argv.remove('--nr-processes={0}'.format(nr_processes))
 
+    vprint_stderr = VerbosityPrinter(options, sys.stderr)
     #
     # Remove all non-option args from argv before remote
     # execution; these values will be trickled down and
@@ -262,6 +313,7 @@ def do_fork(opt, ToolImplementation, tool_reads_stdin, args, nr_processes, nr_mi
         argv.append('-')
 
     subprocesses = []
+    sighandler = SignalHandler()
 
     #
     # Spawn subprocesses.
@@ -270,6 +322,13 @@ def do_fork(opt, ToolImplementation, tool_reads_stdin, args, nr_processes, nr_mi
     for i in xrange(nr_processes):
         proc = subprocess.Popen(argv, stdin=subprocess.PIPE)
         subprocesses.append(proc)
+        #
+        # Register cleanup handlers in case of ^C etc.
+        #
+        cleanup_msg = 'Popen.pid {0}, stdin.close()'.format(proc.pid)
+        sighandler.register_cleanup(0, proc.stdin.close, _msg=cleanup_msg)
+        cleanup_msg = 'Popen.pid {0}, wait()'.format(proc.pid)
+        sighandler.register_cleanup(100, proc.wait, _msg=cleanup_msg)
 
     for (i, item) in enumerate(inputchain):
         #
@@ -284,11 +343,11 @@ def do_fork(opt, ToolImplementation, tool_reads_stdin, args, nr_processes, nr_mi
         proc.stdin.write(item)
 
     for proc in subprocesses:
-        #print os.getpid(), 'CLOSE STDIN', proc
+        vprint_stderr('[*] ii', 'CLOSE-STDIN', os.getpid(), proc)
         proc.stdin.close()
 
     for proc in subprocesses:
-        #print os.getpid(), 'WAIT', proc
+        vprint_stderr('[*] ii', 'WAIT', os.getpid(), proc)
         proc.wait()
 
 
@@ -296,7 +355,7 @@ def run(ToolImplementation, tool_reads_stdin=False):
     try:
         return _run(ToolImplementation, tool_reads_stdin)
     except KeyboardInterrupt:
-        return 1
+        SignalHandler().do_cleanup()
 
 
 def _run(ToolImplementation, tool_reads_stdin):
@@ -316,9 +375,12 @@ def _run(ToolImplementation, tool_reads_stdin):
         itemname = ToolImplementation.__itemname__
         _usage = _usage.replace('item', itemname)
     parser = XNetOptionParser(ToolImplementation.cmdline_options, usage=_usage)
-    (opt, args) = parser.parse_args(args)
+    (options, args) = parser.parse_args(args)
+    SignalHandler.setup(options)
     cmdlineinputs = args[1:]
 
+    #vprint = VerbosityPrinter(options)
+    vprint_stderr = VerbosityPrinter(options, sys.stderr)
     #
     # chain input sources, handle options
     # precedence order: cmdline, -r, stdin
@@ -332,47 +394,47 @@ def _run(ToolImplementation, tool_reads_stdin):
             raise Exception(_E_INVALID_ARGUMENT_STDIN)
         cmdlineinputs.remove('-')
         _inputs.append(sys.stdin)
-    if opt.read:
-        f = open(opt.read)
+    if options.read:
+        f = open(options.read)
         _inputs.insert(0, f)
     if len(cmdlineinputs):
         _inputs.insert(0, cmdlineinputs)
     inputchain = itertools.chain(*_inputs)
 
     _nr_processes = 1
-    if opt.nr_processes:
-        _nr_processes = int(opt.nr_processes)
+    if options.nr_processes:
+        _nr_processes = int(options.nr_processes)
 
-    if opt.pdb:
+    if options.pdb:
         import xnet.debug
         xnet.debug.interactive_debugger_on_exception(True)
 
-    if opt.print_source:
+    if options.print_source:
         ToolImplementation.print_source()
         return
 
     _nr_microthreads = 128
-    if not opt.nr_microthreads is None:
-        _nr_microthreads = int(opt.nr_microthreads)
+    if not options.nr_microthreads is None:
+        _nr_microthreads = int(options.nr_microthreads)
 
     _wait = None
-    if not opt.wait is None:
-        _wait = float(opt.wait)
+    if not options.wait is None:
+        _wait = float(options.wait)
 
     _interval = 0.0
-    if not opt.interval is None:
-        _interval = float(opt.interval)
+    if not options.interval is None:
+        _interval = float(options.interval)
 
     _repeat = 1
-    if not opt.repeat is None:
-        _repeat = int(opt.repeat)
+    if not options.repeat is None:
+        _repeat = int(options.repeat)
 
-    if opt.format_help:
+    if options.format_help:
         print ToolImplementation.__format_help__()
         sys.exit(0)
 
-    if opt.split_tee:
-        opt.split_output = opt.split_tee
+    if options.split_tee:
+        options.split_output = options.split_tee
 
     _outfile = sys.stdout
 
@@ -380,16 +442,16 @@ def _run(ToolImplementation, tool_reads_stdin):
     # Handle SSH-distributed execution.
     # If SSH-dist., main process returns on do_ssh() here.
     #
-    if opt.ssh_nodes_file:
+    if options.ssh_nodes_file:
         #
         # __massage__() expands wildcard IP-ranges such as 10.0.0.*.
         # This should be done in parent in order to split the expanded
         # set of IP:s across its children.
         #
-        inputchain = ToolImplementation.__massage__(inputchain, opt)
+        inputchain = ToolImplementation.__massage__(inputchain, options)
         if _repeat != 1:
             inputchain = repeaterator(inputchain, _repeat)
-        return do_ssh(opt, ToolImplementation, args, inputchain)
+        return do_ssh(options, ToolImplementation, args, inputchain)
 
     #
     # Handle multiproc.
@@ -407,22 +469,22 @@ def _run(ToolImplementation, tool_reads_stdin):
         # This should be done in parent in order to split the expanded
         # set of IP:s across its children.
         #
-        inputchain = ToolImplementation.__massage__(inputchain, opt)
+        inputchain = ToolImplementation.__massage__(inputchain, options)
         if _repeat != 1:
             inputchain = repeaterator(inputchain, _repeat)
-        return do_fork(opt, ToolImplementation, tool_reads_stdin, args,
+        return do_fork(options, ToolImplementation, tool_reads_stdin, args,
                        _nr_processes, _nr_microthreads, inputchain)
 
     #
     # verify commandline options and do preparations
     #
-    ToolImplementation.__setup__(opt)
+    ToolImplementation.__setup__(options)
 
     #
     # Errors if pool is too small.
     #
     pool = gevent.pool.Pool(_nr_microthreads)
-    #pool = gevent.pool.Pool(opt.nr_microthreads or DEFAULT_NR_MICROTHREADS)
+    #pool = gevent.pool.Pool(options.nr_microthreads or DEFAULT_NR_MICROTHREADS)
     greenlets = []
 
     wkpool = None
@@ -432,14 +494,22 @@ def _run(ToolImplementation, tool_reads_stdin):
     def waitkill(g, killtime):
         from xnet.tools import WaitTimeout
         sleeptime = killtime - time.time()
-        #print 'waitkill', sleeptime
         if sleeptime > 0:
             gevent.sleep(sleeptime)
         if not g.ready():
+            vprint_stderr(3, '[*] xx - kill {0}:{1}\n'.format(
+                os.getpid(), g.action._greenlet_id
+            ))
             gevent.kill(g, WaitTimeout)
+            tstart = time.time()
             g.join()
+            tdiff = time.time() - tstart
+            vprint_stderr(3, '[*] yy - joined {0}:{1} ({2:.1f}s)\n'.format(
+                os.getpid(), g.action._greenlet_id, tdiff
+            ))
+            assert g.ready()
 
-    inputchain = ToolImplementation.__massage__(inputchain, opt)
+    inputchain = ToolImplementation.__massage__(inputchain, options)
 
     kwargs = {}
 
@@ -454,11 +524,12 @@ def _run(ToolImplementation, tool_reads_stdin):
 
     greenlet_id = 0
     for (i, line) in enumerate(inputchain):
-        action = ToolImplementation(opt, greenlet_id=greenlet_id, **kwargs)
+        action = ToolImplementation(options, greenlet_id=greenlet_id, **kwargs)
         greenlet_id += 1
         pool.wait_available()
         if _wait:
             killtime = time.time() + _wait
+        vprint_stderr(2, '[*] ++ spawning greenlet {0}:{1}\n'.format(os.getpid(), greenlet_id))
         g = pool.spawn(action, line, inputchain)
         g.action = action
         greenlets.append(g)
@@ -472,11 +543,32 @@ def _run(ToolImplementation, tool_reads_stdin):
         #
         # handle finished actions
         #
-        while len(greenlets) and greenlets[0].ready():
-            action = greenlets[0].action
-            output_action_result(action, opt, _outfile)
-            del action
-            del greenlets[0]
+        vprint_stderr(3, '[*] ii len(greenlets) = {0}'.format(len(greenlets)))
+        vprint_stderr(3, ', not_ready={0}\n'.format(len(filter(
+            lambda g: not g.ready(), greenlets
+        ))))
+
+        #while len(greenlets) and greenlets[0].ready():
+        #    action = greenlets[0].action
+        #    vprint_stderr(2, '[*] -- collecting greenlet: {0}:{1} (running: {2})\n'.format(
+        #        os.getpid(), action._greenlet_id, len(greenlets)
+        #    ))
+        #    output_action_result(action, options, _outfile)
+        #    del action
+        #    del greenlets[0]
+
+        del_indexes = []
+        for (g_index, g) in enumerate(greenlets[:]):
+            if not g.ready():
+                continue
+            vprint_stderr(2, '[*] -- collecting greenlet: {0}:{1} (running: {2})\n'.format(
+                os.getpid(), g.action._greenlet_id, len(greenlets)
+            ))
+            output_action_result(g.action, options, _outfile)
+            del_indexes.append(g_index)
+        while len(del_indexes):
+            di = del_indexes.pop()
+            del greenlets[di]
 
         #
         # handle interval
@@ -488,11 +580,11 @@ def _run(ToolImplementation, tool_reads_stdin):
 
     if wkpool:
         wkpool.join()
-    gevent.joinall(greenlets)  #, timeout=timeout)
+    gevent.joinall(greenlets)  # timeout=timeout
     not_done = filter(lambda g: (not g.ready()), greenlets)
     if len(not_done) > 0 and not _wait is None:
-        print 'ERROR: not_done has contents inspite of _wait and grace time'
-        print not_done
+        vprint_stderr(0, 'ERROR: not_done has contents inspite of _wait and grace time\n')
+        vprint_stderr(0,  not_done)
 
     #
     # Force-kill greenlets that didn't die in spite of 1 sec of grace time.
@@ -503,14 +595,14 @@ def _run(ToolImplementation, tool_reads_stdin):
     #
     # cleanups
     #
-    ToolImplementation.__teardown__(opt)
+    ToolImplementation.__teardown__(options)
 
     #
     # print results
     #
     for g in greenlets:
         action = g.action
-        output_action_result(action, opt, _outfile)
+        output_action_result(action, options, _outfile)
         del action
 
 
